from __future__ import annotations

import sys
from typing import List, Tuple


class WizardBase:
    steps: List["WizardBase"] = []
    parent: "WizardBase" = None

    @property
    def name(self):
        return type(self).__name__

    @property
    def root(self):
        n = self
        while n.parent:
            n = n.parent
        return n

    @property
    def data(self):
        if self.root == self:
            raise RuntimeError(
                "Parent not initialized, should not access data at this time!"
            )
        return self.root.data

    def vprint(self, *args, **kwargs):
        if self.root.verbose:
            print(*args, **kwargs, file=sys.__stderr__)

    def __init__(self, steps: "Tuple[WizardBase]" = None):
        self.steps = steps or self.steps
        for step in self.steps:
            step.parent = self

    def run(self, selected_steps=None):
        steps = self.steps
        self.vprint(f"{self.name} running steps")
        for stepidx, step in enumerate(steps):
            if selected_steps and step.name not in selected_steps:
                continue
            self.vprint(f"Checking if step {step.name} should be run")
            should_run = step.should_run()
            if should_run is False:
                self.vprint(f"   No, skipping")
                continue
            if should_run is None:
                # only if one of the subsequent steps should run
                for subsequent in steps[stepidx + 1 :]:
                    subsequent_should_run = subsequent.should_run()
                    if subsequent_should_run is not None:
                        should_run = subsequent_should_run
                        break
                if should_run is False:
                    continue
            self.vprint(f"Calling step {step.name}")
            step.run()
            self.data.save()

    def should_run(self):
        if self.steps:
            for step in self.steps:
                if isinstance(step, str):
                    getattr(self, step)()
                should_run = step.should_run()
                if should_run:
                    return True
            return False



from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.site.site_support import SiteSupport


class InstallModelStep(ModelWizardStep):
    def should_run(self):
        return True

    def after_run(self):
        sites = self.data["sites"]
        for site in sites:
            SiteSupport(self.data, site).rebuild_site()



from oarepo_cli.model.gen.base import GeneratedFiles
from oarepo_cli.wizard import StaticStep, Wizard

from .steps.add_custom_fields import AddCustomFieldsWizardStep
from .steps.add_docs_base import AddDocsBaseWizardStep
from .steps.add_files import AddFilesWizardStep
from .steps.add_metadata import AddMetadataWizardStep
from .steps.add_nr_vocabularies import AddNRVocabulariesWizardStep
from .steps.add_permissions import AddPermissionsWizardStep
from .steps.add_pid_type import AddPIDTypeWizardStep
from .steps.add_relations import AddRelationsWizardStep
from .steps.add_tests import AddTestsWizardStep
from .steps.add_vocabularies import AddVocabulariesWizardStep
from .steps.create_model import EmptyModelWizardStep


class AddModelWizard(Wizard):
    def __init__(self):
        super().__init__(
            StaticStep(
                heading="""
        Before creating the datamodel, I'll ask you a few questions.
        If unsure, use the default value.
            """,
            ),
            # these just generate sources, no need to run them inside docker
            EmptyModelWizardStep(),
            AddDocsBaseWizardStep(),
            AddFilesWizardStep(),
            AddCustomFieldsWizardStep(),
            AddNRVocabulariesWizardStep(),
            AddVocabulariesWizardStep(),
            AddRelationsWizardStep(),
            AddMetadataWizardStep(),
            AddPermissionsWizardStep(),
            AddPIDTypeWizardStep(),
            AddTestsWizardStep(),
        )

    def run_wizard(
        self, data, *, no_input=False, silent=False, selected_steps=None, verbose=False
    ):
        self.files = GeneratedFiles(data.project_dir / "models" / data.section)
        super().run_wizard(
            data,
            no_input=no_input,
            silent=silent,
            selected_steps=selected_steps,
            verbose=verbose,
        )



from oarepo_cli.develop.steps.check_site import CheckSiteStep
from oarepo_cli.develop.steps.check_ui import CheckUIStep
from oarepo_cli.develop.steps.check_venv import CheckVirtualenvStep
from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.wizard import Wizard
from oarepo_cli.wizard.docker import DockerRunner


class BuildWizard(Wizard):
    def __init__(self, runner: DockerRunner, site_support: SiteSupport, production: bool):

        super().__init__(
            *runner.wrap_docker_steps(
                CheckVirtualenvStep(clean=True),
                CheckSiteStep(clean=True),
                CheckUIStep(production=production)
            ),
        )



from __future__ import annotations

import re

from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.utils import run_cmdline
from oarepo_cli.wizard import WizardStep


class CheckRequirementsStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
I will check the requirements for Invenio site installation.
            """,
            **kwargs,
        )

    def after_run(self):
        # no check for python performed
        self._check_pdm_callable()
        self._check_docker_callable()
        self._check_docker_compose_version(1, 17)
        self._check_node_version(14, 16)
        self._check_npm_version(6, 7, 8)
        self._check_imagemagick_callable()
        with open(self.site_dir / ".check.ok", "w") as f:
            f.write("invenio check ok")

    def should_run(self):
        return not (self.site_dir / ".check.ok").exists()

    def _check_docker_callable(self):
        if self.data.running_in_docker:
            return
        run_cmdline("docker", "ps", grab_stdout=True)

    def _check_pdm_callable(self):
        run_cmdline(
            "pdm",
            "--version",
            environ={"PDM_IGNORE_ACTIVE_VENV": "1"},
            grab_stdout=True,
        )

    def _check_imagemagick_callable(self):
        run_cmdline("convert", "--version", grab_stdout=True)

    def check_version(self, *args, expected_major, expected_minor=None, strict=False):
        result = run_cmdline(*args, grab_stdout=True)
        self.vprint(f"Version string is: {result}")
        version_result = re.search(r".*?([0-9]+)\.([0-9]+)\.([0-9]+)", result)
        major = int(version_result.groups()[0])
        minor = int(version_result.groups()[1])
        if strict:
            if isinstance(expected_major, (list, tuple)):
                assert (
                    major in expected_major
                ), f"Expected major version to be one of {expected_major}, found {major}"
            else:
                assert (
                    major == expected_major
                ), f"Expected major version to be one {expected_major}, found {major}"
                if expected_minor:
                    assert minor == expected_minor
        elif not (
            major > expected_major
            or (major == expected_major and minor >= expected_minor)
        ):
            raise Exception("Expected docker compose version ")

    def _check_docker_compose_version(self, expected_major, expected_minor):
        if self.data.running_in_docker:
            return
        self.check_version(
            "docker-compose",
            "--version",
            expected_major=expected_major,
            expected_minor=expected_minor,
        )

    def _check_node_version(self, *supported_versions):
        self.check_version(
            "node", "--version", expected_major=supported_versions, strict=True
        )

    def _check_npm_version(self, *supported_versions):
        self.check_version(
            "npm", "--version", expected_major=supported_versions, strict=True
        )



import shutil
from pathlib import Path
from typing import List

import click as click
import yaml

from oarepo_cli.utils import commit_git, with_config

from ..model_support import ModelSupport
from .wizard import AddModelWizard


@click.command(
    name="add",
    help="""
Generate a new model. Required arguments:
    <name>   ... name of the model, can contain [a-z] and dash (-)""",
)
@click.argument("name", required=True)
@click.option(
    "--use",
    multiple=True,
    help="""
Use this option to merge your code into the generated model.

Syntax: --use <custom_model_file_relative_to_cwd>
or      --use <custom_model_file_relative_to_cwd>:<jsonpath>

The file will be copied to the destination and referenced from 
the generated model file. If no path is specified, it will be 
referenced from the root of the file, with path the reference 
will be put there. Only '/' is supported in the json path.
""",
)
@with_config(config_section=lambda name, **kwargs: ["models", name])
def add_model(
    cfg=None,
    use=None,
    step=None,
    no_input=False,
    silent=False,
    verbose=False,
    steps=False,
    **kwargs,
):
    commit_git(
        cfg.project_dir,
        f"before-model-add-{cfg.section}",
        f"Committed automatically before model {cfg.section} has been added",
    )
    cfg["model_dir"] = f"models/{cfg.section}"

    wizard = AddModelWizard()
    if steps:
        wizard.list_steps()
        return

    wizard.run_wizard(
        cfg, selected_steps=step, no_input=no_input, silent=silent, verbose=verbose
    )
    model_support = ModelSupport(cfg)

    merge_extra_sources(model_support, use or [])
    commit_git(
        cfg.project_dir,
        f"after-model-add-{cfg.section}",
        f"Committed automatically after model {cfg.section} has been added",
    )


def merge_extra_sources(model_support: ModelSupport, sources: List[str]):
    for d in sources:
        d = d.split(":")
        if len(d) < 2:
            filepath = d[0]
            jsonpath = "/"
        else:
            filepath, jsonpath = d

        extra_file_name = Path(filepath).name
        model_dir = model_support.model_dir
        model_file_path = model_dir / "model.yaml"

        shutil.copy(filepath, model_dir / extra_file_name)

        with model_file_path.open() as f:
            json_data = yaml.safe_load(f)

        add_oarepo_use(json_data, jsonpath.split("/"), f"./{extra_file_name}")

        with model_file_path.open("wt") as f:
            yaml.safe_dump(json_data, f)


def add_oarepo_use(d, path, value):
    for p in path:
        if not p:
            continue
        if p not in d:
            d[p] = {}
        d = d[p]

    if "use" in d:
        if not isinstance(d["use"], list):
            d["use"] = [d["use"]]
    else:
        d["use"] = []
    d["use"].append(value)



import click

from oarepo_cli.utils import commit_git, to_python_name, with_config
from oarepo_cli.wizard.docker import DockerRunner

from ..site_support import SiteSupport
from .wizard import AddSiteWizard


@click.command(
    name="add",
    help="""Generate a new site.  Required arguments:
    <name>   ... name of the site. The recommended pattern for it is <something>-site""",
)
@click.argument("name")
@with_config(config_section=lambda name, **kwargs: ["sites", name])
def add_site(
    cfg=None,
    name=None,
    no_input=False,
    silent=False,
    step=None,
    verbose=False,
    steps=False,
    **kwargs,
):
    commit_git(
        cfg.project_dir,
        f"before-site-install-{cfg.section}",
        f"Committed automatically before site {cfg.section} has been added",
    )
    cfg["site_package"] = to_python_name(name)
    cfg["site_dir"] = f"sites/{name}"
    site_support = SiteSupport(cfg)
    runner = DockerRunner(cfg, no_input)
    initialize_wizard = AddSiteWizard(runner, site_support)
    if steps:
        initialize_wizard.list_steps()
        return
    cfg["pdm_name"] = name
    initialize_wizard.run_wizard(
        cfg, no_input=no_input, silent=silent, selected_steps=step, verbose=verbose
    )
    commit_git(
        cfg.project_dir,
        f"after-site-install-{cfg.section}",
        f"Committed automatically after site {cfg.section} has been added",
    )



import click as click

from oarepo_cli.utils import with_config

from ...site.site_support import SiteSupport
from ...wizard.docker import DockerRunner
from .wizard import InstallWizard


@click.command(
    name="install",
    help="""
    Install the UI to the site. Required arguments:
    <name>   ... name of the ui. The recommended pattern for it is <modelname>-ui
    """,
)
@click.argument("name")
@with_config(config_section=lambda name, **kwargs: ["ui", name])
def install_ui(
    cfg=None,
    step=None,
    no_input=False,
    silent=False,
    verbose=False,
    steps=False,
    **kwargs
):
    site_support = SiteSupport(cfg)
    runner = DockerRunner(cfg, no_input)

    wizard = InstallWizard(runner, site_support=site_support)
    if steps:
        wizard.list_steps()
        return

    wizard.run_wizard(
        cfg, selected_steps=step, no_input=no_input, silent=silent, verbose=verbose
    )



from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import WizardStep


class CheckUIStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, production=False):
        super().__init__()
        self.production = production

    def after_run(self):
        self.site_support.build_ui(production=self.production)

    def should_run(self):
        # production always rebuilds the site
        return self.production or not self.site_support.ui_ok()



import shutil
import venv

from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.utils import run_cmdline
from oarepo_cli.wizard import WizardStep


class RunTestsModelStep(ModelWizardStep, WizardStep):
    def after_run(self):
        if self.data["run_tests"] == "skip":
            return
        model_dir = self.model_dir
        venv_dir = model_dir / ".venv-test"
        if venv_dir.exists():
            shutil.rmtree(venv_dir)

        venv.main([str(venv_dir)])
        pip_binary = venv_dir / "bin" / "pip"
        pytest_binary = venv_dir / "bin" / "pytest"

        run_cmdline(
            pip_binary, "install", "-U", "--no-input", "setuptools", "pip", "wheel"
        )
        run_cmdline(
            pip_binary, "install", "--no-input", "-e", ".[tests]", cwd=model_dir
        )

        run_cmdline(
            pytest_binary,
            "tests",
            cwd=model_dir,
            environ={
                "OPENSEARCH_HOST": self.data.get(
                    "config.TEST_OPENSEARCH_HOST", "localhost"
                ),
                "OPENSEARCH_PORT": self.data.get("config.TEST_OPENSEARCH_PORT", "9400"),
            },
        )

    def should_run(self):
        return True



from typing import List

import click as click

from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import commit_git, with_config
from oarepo_cli.wizard.docker import DockerRunner

from .wizard import RemoveLocalWizard


@click.command(
    name="remove",
    help="""Remove a local package:
    <name>               ... pypi name of the package
    """,
)
@click.argument("name")
@click.argument("site_name", required=False)
@with_config(config_section=lambda name, **kwargs: ["local", name])
def remove_local(
    cfg=None,
    step=None,
    no_input=False,
    silent=False,
    verbose=False,
    steps=False,
    site_name=None,
    **kwargs,
):
    commit_git(
        cfg.project_dir,
        f"before-local-remove-{cfg.section}",
        f"Committed automatically before package {cfg.section} has been removed",
    )

    site_support = SiteSupport(cfg, site_name)

    list_sites: List[str] = cfg.setdefault("sites", [])
    list_sites.remove(site_support.site_name)
    cfg.save()

    runner = DockerRunner(cfg, no_input)
    wizard = RemoveLocalWizard(runner, site_support=site_support)
    if steps:
        wizard.list_steps()
        return

    wizard.run_wizard(
        cfg, selected_steps=step, no_input=no_input, silent=silent, verbose=verbose
    )

    commit_git(
        cfg.project_dir,
        f"after-local-clone-{cfg.section}",
        f"Committed automatically after package {cfg.section} has been removed",
    )



from oarepo_cli.wizard import Wizard

from ...site.site_support import SiteSupport
from ...wizard.docker import DockerRunner
from .steps.add_local import GitHubCloneWizardStep
from .steps.install import InstallToSiteStep


class AddLocalWizard(Wizard):
    def __init__(self, runner: DockerRunner, site_support: SiteSupport):
        self.site_support = site_support

        super().__init__(
            GitHubCloneWizardStep(),
            *runner.wrap_docker_steps(InstallToSiteStep()),
        )



import errno
import os
import time

import click as click

from oarepo_cli.develop.config import CONTROL_PIPE
from oarepo_cli.develop.wizard import DevelopWizard
from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import with_config
from oarepo_cli.wizard.docker import DockerRunner


@click.command(
    name="develop",
    hidden=True,
    help="Use this command to start development server (either in docker or in userspace)",
)
@click.option("--site", required=False)
@click.option("--command", required=False, hidden=True)
@with_config()
def develop_command(
    cfg,
    no_input=False,
    silent=False,
    step=None,
    verbose=False,
    steps=False,
    site=None,
    command=None,
    **kwargs
):
    if command:
        # there is a CONTROL_PIPE pipe, send command to it and quit
        send_command(command)
        return

    site_support = SiteSupport(cfg, site)

    runner = DockerRunner(cfg, no_input)
    develop_wizard = DevelopWizard(runner, site_support=site_support)
    if steps:
        develop_wizard.list_steps()
        return

    develop_wizard.run_wizard(
        cfg, no_input=no_input, silent=silent, selected_steps=step, verbose=verbose
    )


def send_command(command):
    # make sure CONTROL_PIPE is a pipe
    try:
        os.mkfifo(CONTROL_PIPE)
    except OSError as oe:
        if oe.errno != errno.EEXIST:
            raise

    with open(CONTROL_PIPE, "w") as f:
        f.write(command + "\n")
        f.flush()
        time.sleep(2)



from __future__ import annotations

import readline

from colorama import Fore, Style

from .base import WizardBase


class WizardStep(WizardBase):
    heading = ""
    pause = None

    def __init__(
        self,
        *steps: "WizardBase",
        heading=None,
        pause=False,
        **kwargs,
    ):
        super().__init__(steps)
        self.heading = heading or self.heading
        self.pause = pause or self.pause

    def run(self, selected_steps=None):
        if not self.on_before_run():
            return
        self.on_before_heading()
        self._print_heading()
        self.on_after_heading()
        super().run(selected_steps=selected_steps)
        if self.pause and not self.root.no_input:
            input(f"Press enter to continue ...")
        self.after_run()

    def _print_heading(self):
        if self.heading and not self.root.silent:
            heading = self.heading
            if callable(heading):
                heading = heading(self.data)
            if heading:
                print(f"\n\n{Fore.BLUE}{heading.strip()}{Style.RESET_ALL}")
            print()

    def on_before_run(self):
        return True

    def on_before_heading(self):
        pass

    def on_after_heading(self):
        pass

    def on_after_steps(self):
        pass

    def after_run(self):
        pass


class InputStep(WizardStep):
    def __init__(
        self,
        key,
        heading=None,
        required=True,
        default=None,
        prompt=None,
        force_run=False,
    ):
        super().__init__(
            heading=heading,
        )
        self.key = key
        self.force_run = force_run
        self._default = default
        self.prompt = prompt
        self.required = required

    @property
    def default(self):
        if callable(self._default):
            return self._default(self.data)
        return self._default

    def run(self, **kwargs):
        value = self.data.get(self.key, self.default)

        while True:
            try:
                prompt = self.prompt or "Enter value"
                if self.default:
                    prompt += f" [{self.default}]"
                line = input(f"{Fore.BLUE}{prompt}: {Style.RESET_ALL}")
                line = line.strip() or value
                if line or not self.required:
                    self.data[self.key] = line
                    return
            finally:
                readline.set_pre_input_hook()

    def should_run(self):
        return self.force_run or self.key not in self.data


class StaticStep(WizardStep):
    def __init__(self, heading, **kwargs):
        super().__init__(heading=heading, **kwargs)

    def should_run(self):
        # do not know - should run only if one of the subsequent steps should run
        return None


class RadioStep(WizardStep):
    def __init__(self, key, heading=None, options=None, default=None, force_run=False):
        super().__init__(heading=heading)
        self.key = key
        self.force_run = force_run
        self._default = default
        self.options = options

    @property
    def default(self):
        if callable(self._default):
            return self._default(self.data)
        return self._default

    def run(self, **kwargs):
        super().run(**kwargs)
        value = self.data.get(self.key, self.default)

        options = self.options
        if callable(options):
            options = options()

        displayed = [
            (str(idx + 1), key, label)
            for idx, (key, label) in enumerate(options.items())
        ]
        print()
        for d in displayed:
            print(f"{Fore.YELLOW}{d[0]}{Style.RESET_ALL}) {d[2]}")

        option = None
        for d in displayed:
            if d[1] == value:
                option = d[0]
                break
        if option:
            prompt = f"Your choice [{option}]: "
        else:
            prompt = "Your choice: "

        while True:
            print()
            value = input(prompt).strip() or option
            for d in displayed:
                if d[0] == value:
                    self.data[self.key] = d[1]
                    return
            print(
                f"{Fore.RED}Bad option: select one of the options above{Style.RESET_ALL}"
            )

    def should_run(self):
        return (not self.data.no_input and self.force_run) or self.key not in self.data



from invenio_assets.webpack import WebpackThemeBundle

theme = WebpackThemeBundle(
    __name__,
    "assets",
    default="semantic-ui",
    themes={
        "semantic-ui": dict(
            entry={
                "{{cookiecutter.app_package}}_components": "./js/{{cookiecutter.app_package}}/custom-components.js",
                "{{cookiecutter.app_package}}_search": "./js/{{cookiecutter.app_package}}/search/index.js",
            },
            dependencies={
                "react-searchkit": "^2.0.0",
            },
            devDependencies={},
            aliases={
                "@translations/{{cookiecutter.app_package}}": "translations/{{cookiecutter.app_package}}",
            },
        )
    },
)



from oarepo_cli.initialize.steps.create_monorepo import CreateMonorepoStep
from oarepo_cli.initialize.steps.initialize_directory import MonorepoDirectoryStep
from oarepo_cli.initialize.steps.install_nrp_cli import InstallINRPCliStep
from oarepo_cli.wizard import StaticStep, Wizard


class InitializeWizard(Wizard):
    def __init__(self):
        super().__init__(
            StaticStep(
                """
            This command will initialize a new repository based on OARepo codebase (an extension of Invenio repository).
                    """,
                pause=True,
            ),
            MonorepoDirectoryStep(),
            CreateMonorepoStep(),
            InstallINRPCliStep(pause=True),
        )



NRP_CLI_VERSION = "oarepo-cli>=11.0.8,<12"
OAREPO_VERSION = "oarepo>=11,<12"
PYTHON_VERSION = ">=3.9,<3.11"

NR_METADATA_VERSION = ">=1.0.0,<2"
NR_OAIPMH_HARVESTERS_VERSION = ">=1.0.0,<2"
NR_VOCABULARIES_VERSION = ">=1.0.0,<2"

OAREPO_MODEL_BUILDER_CF_VERSION = "==4.*"
OAREPO_MODEL_BUILDER_FILES_VERSION = "==4.*"
OAREPO_MODEL_BUILDER_NR_VERSION = "==4.*"
OAREPO_MODEL_BUILDER_RELATIONS_VERSION = "==4.*"
OAREPO_MODEL_BUILDER_VOCABULARIES_VERSION = "==4.*"
OAREPO_MODEL_BUILDER_TESTS_VERSION = "==4.*"
OAREPO_MODEL_BUILDER_VERSION = "==4.*"



from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import WizardStep


class InstallInvenioStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, clean=False, **kwargs):
        super().__init__(
            heading="""
Now I'll install invenio site.
            """,
            **kwargs,
        )
        self.clean = clean

    def after_run(self):
        self.site_support.check_and_create_virtualenv(clean=self.clean)

        self.site_support.install_site()

    def should_run(self):
        return True



import click as click

from oarepo_cli.utils import with_config


@click.command(name="list", help="List installed sites")
@with_config()
def list_sites(cfg=None, **kwargs):
    for site in cfg.whole_data.get("sites", {}):
        print(site)



from oarepo_ui.resources.config import RecordsUIResourceConfig


class {{cookiecutter.resource_config}}(RecordsUIResourceConfig):
    template_folder = "../templates"
    url_prefix = "{{cookiecutter.url_prefix}}"
    blueprint_name = "{{cookiecutter.app_name}}"
    ui_serializer_class = "{{cookiecutter.ui_serializer_class}}"
    api_service = "{{cookiecutter.api_service}}"
    layout = "{{cookiecutter.api_service}}"

    templates = {
        "detail": {
            "layout": "{{cookiecutter.app_package}}/detail.html",
            "blocks": {
                "record_main_content": "{{cookiecutter.app_package}}/main.html",
                "record_sidebar": "{{cookiecutter.app_package}}/sidebar.html"                
            },
        },
        "search": {"layout": "{{cookiecutter.app_package}}/search.html"},
    }



# Copyright (c) 2022 CESNET
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from .ext import {{ cookiecutter.ext_name }}

__version__ = '1.0.0'

__all__ = ('__version__', '{{cookiecutter.ext_name}}')



from os.path import relpath

from oarepo_cli.templates import get_cookiecutter_template
from oarepo_cli.ui.add.mixins import AssociatedModelMixin, UIWizardMixin
from oarepo_cli.utils import ProjectWizardMixin, snail_to_title, to_python_name
from oarepo_cli.wizard import InputStep, RadioStep, WizardStep


class AddUIWizardStep(
    AssociatedModelMixin, UIWizardMixin, ProjectWizardMixin, WizardStep
):
    def __init__(self):
        super().__init__(
            RadioStep(
                "model_name",
                heading="""
                For which model do you want to generate the ui?
                """,
                options=self.available_models,
                default=lambda data: next(iter(self.available_models())),
            ),
            InputStep(
                "url_prefix",
                prompt="On which url prefix will the UI reside? The prefix should like /something/: ",
                default=lambda data: f"/{data.section}/",
            ),
        )

    def available_models(self):
        known_models = {x: x for x in self.data.whole_data.get("models", {}).keys()}
        return known_models

    def available_sites(self):
        return list(self.data.whole_data.get("sites", {}).keys())

    def after_run(self):
        ui_name = self.ui_name

        ui_package = to_python_name(ui_name)
        ui_base = snail_to_title(ui_package)

        (
            model_description,
            model_path,
            model_package,
            model_config,
        ) = self.get_model_definition()

        if not self.data["url_prefix"].startswith("/"):
            self.data["url_prefix"] = "/" + self.data["url_prefix"]
        if not self.data["url_prefix"].endswith("/"):
            self.data["url_prefix"] += "/"

        model_service = model_description["model"]["service-config"]["service-id"]
        ui_serializer_class = model_description["model"]["json-serializer"]["class"]

        self.data.setdefault("sites", self.available_sites())

        self.data.setdefault(
            "cookiecutter_local_model_path", relpath(model_path, self.ui_dir)
        )
        self.data.setdefault("cookiecutter_model_package", model_package)
        self.data.setdefault("cookiecutter_app_name", ui_name)
        self.data.setdefault("cookiecutter_app_package", ui_package)
        self.data.setdefault("cookiecutter_ext_name", f"{ui_base}Extension")

        self.data.setdefault("cookiecutter_author", model_config.get("author_name", ""))
        self.data.setdefault(
            "cookiecutter_author_email", model_config.get("author_email", "")
        )
        self.data.setdefault("cookiecutter_repository_url", "")
        # TODO: take this dynamically from the running model's Ext so that
        # TODO: it does not have to be specified here
        self.data.setdefault("cookiecutter_resource", f"{ui_base}Resource")
        self.data.setdefault("cookiecutter_resource_config", f"{ui_base}ResourceConfig")
        self.data.setdefault("cookiecutter_api_service", model_service)
        self.data.setdefault(
            "cookiecutter_ui_record_serializer_class", ui_serializer_class
        )

        cookiecutter_data = {
            "model_name": self.data["model_name"],
            "local_model_path": self.data["cookiecutter_local_model_path"],
            "model_package": self.data["cookiecutter_model_package"],
            "app_name": self.data["cookiecutter_app_name"],
            "app_package": self.data["cookiecutter_app_package"],
            "ext_name": self.data["cookiecutter_ext_name"],
            "author": self.data["cookiecutter_author"],
            "author_email": self.data["cookiecutter_author_email"],
            "repository_url": self.data["cookiecutter_repository_url"],
            # TODO: take this dynamically from the running model's Ext so that
            # TODO: it does not have to be specified here
            "resource": self.data["cookiecutter_resource"],
            "resource_config": self.data["cookiecutter_resource_config"],
            "api_service": self.data["cookiecutter_api_service"],
            "ui_serializer_class": self.data["cookiecutter_ui_record_serializer_class"],
            "url_prefix": self.data["url_prefix"],
        }

        self.run_cookiecutter(
            template=get_cookiecutter_template("ui"),
            config_file=f"ui-{ui_name}",
            output_dir=self.data.project_dir / "ui",
            extra_context=cookiecutter_data,
        )

        self.data["ui_dir"] = f"ui/{ui_name}"

    def should_run(self):
        return not self.ui_dir.exists()



from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import ProjectWizardMixin, SiteMixin
from oarepo_cli.wizard import WizardStep


class InstallUIStep(SiteMixin, ProjectWizardMixin, WizardStep):
    def should_run(self):
        return True

    def after_run(self):
        sites = self.data["sites"]
        for site in sites:
            SiteSupport(self.data, site).rebuild_site(build_ui=True)



from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import WizardStep


class ResolveDependenciesStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
I am going to resolve python dependencies.
            """,
            **kwargs,
        )

    def after_run(self):
        self.site_support.build_dependencies()

    def should_run(self):
        return True



from oarepo_cli.model.gen.base import GeneratedFile
from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.package_versions import NR_VOCABULARIES_VERSION
from oarepo_cli.utils import unique_merger
from oarepo_cli.wizard import RadioStep


class AddNRVocabulariesWizardStep(ModelWizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            RadioStep(
                "use_nr_vocabularies",
                heading="Do you want your model to use Czech NR vocabularies?",
                options={"yes": "Yes, I want to use Czech NR", "no": "No"},
            ),
            **kwargs
        )

    def after_run(self):
        if self.data["use_nr_vocabularies"] != "yes":
            return
        self.data["use_vocabularies"] = "yes"

        yaml_file: GeneratedFile = self.root.files.get("model.yaml")
        yaml = yaml_file.yaml
        unique_merger.merge(
            yaml,
            {
                "runtime-dependencies": {
                    "nr-vocabularies": NR_VOCABULARIES_VERSION,
                }
            },
        )
        yaml_file.save()

    def should_run(self):
        return True



import shutil

from colorama import Fore, Style

from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.wizard import RadioStep


class RemovePreviousModelStep(ModelWizardStep):
    def __init__(self):
        super().__init__(
            RadioStep(
                "merge_changes",
                options={
                    "merge": "Merge changes into the previously generated files",
                    "overwrite": "Remove previously generated files and start from scratch",
                },
                default="merge",
                heading=f"""
It seems that the model has been already generated. 

Should I try to {Fore.GREEN}merge{Fore.BLUE} the changes with the existing sources 
or {Fore.RED}remove{Fore.BLUE} the previously generated sources and generate from scratch?

{Fore.YELLOW}Please make sure that you have your existing sources safely committed into git repository
so that you might recover them if the compilation process fails.{Style.RESET_ALL}
""",
            )
        )

    def on_before_run(self):
        return (self.model_dir / "setup.cfg").exists()

    def after_run(self):
        if self.data.get("merge_changes") == "overwrite":

            def _rm(x):
                if x.exists():
                    if x.is_dir():
                        shutil.rmtree(x)
                    else:
                        x.unlink()

            _rm(self.model_package_dir)
            _rm(self.model_dir / "setup.cfg")
            _rm(self.model_dir / "data")



import shutil

from oarepo_cli.format.steps.base import BaseFormatStep
from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import batched, run_cmdline


class FormatJavascriptStep(BaseFormatStep):
    def format_paths(self, model_paths, ui_paths, local_paths, site_paths):
        self.format_jsx(ui_paths + local_paths + site_paths, [])

    def format_jsx(self, dirs, exclude):
        site = SiteSupport(self.data)
        assets_dir = site.invenio_instance_path / "assets"
        if assets_dir.exists() and (assets_dir / "package.json").exists():
            # TODO: only when needed
            run_cmdline("npm", "install", cwd=assets_dir)
        else:
            return

        files = [
            f for f in self.find_files_with_extensions(dirs, "jsx", "tsx", "js", "ts")
        ]
        if not files:
            return
        # copy .eslintrc.json into invenio
        shutil.copy(
            site.site_dir / ".eslintrc.json",
            site.invenio_instance_path / "assets" / ".eslintrc.json",
        )
        for chunk in batched(files, 50):
            run_cmdline(
                site.invenio_instance_path
                / "assets"
                / "node_modules"
                / ".bin"
                / "eslint",
                "-c",
                ".eslintrc.json",
                "--fix",
                *chunk,
                cwd=assets_dir
            )
            run_cmdline(
                site.invenio_instance_path
                / "assets"
                / "node_modules"
                / ".bin"
                / "prettier",
                "--write",
                *chunk,
                cwd=assets_dir
            )



import subprocess
import time
import traceback
from pathlib import Path

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from oarepo_cli.site.assets import copy_watched_paths, load_watched_paths
from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import copy_tree
from oarepo_cli.wizard import WizardStep


class WatcherStep(WizardStep):
    def __init__(self, run_ui):
        super().__init__()
        self.run_ui = run_ui

    def should_run(self):
        return True

    @property
    def site_support(self) -> SiteSupport:
        return self.root.site_support

    @property
    def invenio_instance_path(self):
        return self.site_support.invenio_instance_path

    def after_run(self):
        observer = self.create_observer()
        ui_builder = self.create_ui_builder()

        observer.start()

        while True:
            try:
                # TODO: better termination handling
                time.sleep(60)
            except InterruptedError:
                break

        observer.join()
        if ui_builder:
            try:
                ui_builder.terminate()
                time.sleep(1)
                ui_builder.kill()
            except Exception as e:
                pass

    def create_observer(self):
        destination = self.invenio_instance_path
        watched_paths = load_watched_paths(
            self.invenio_instance_path / "watch.list.json",
            [
                f"{self.site_support.site_dir}/assets=assets",
                f"{self.site_support.site_dir}/static=static",
            ],
        )
        # recursively copy to destination
        copy_watched_paths(watched_paths, destination)
        # start watching
        observer = Observer()
        for path, target in watched_paths.items():
            print(f"Will watch {path} => {target}")
            observer.schedule(
                EventHandler(path, destination / target), path, recursive=True
            )
        return observer

    def create_ui_builder(self):
        if self.run_ui:
            return subprocess.Popen(
                ["npm", "run", "start"], cwd=f"{self.invenio_instance_path}/assets"
            )


class EventHandler(FileSystemEventHandler):
    def __init__(self, source, destination):
        self.source = Path(source)
        self.destination = Path(destination)

    def on_modified(self, event):
        if event.is_directory:
            return

        self._copy(event.src_path)

    def on_moved(self, event):
        self._copy(event.dest_path)

    def _copy(self, path):
        try:
            relative_path = Path(path).relative_to(self.source)
            copy_tree(path, self.destination / relative_path)
        except:
            traceback.print_exc()



import click as click

from .add.cli import add_ui
from .install.cli import install_ui
from .list import list_uis
from .uninstall.cli import uninstall_ui


@click.group(
    help="User interface related tools (add user interface for a model, ...)", name="ui"
)
def ui_commands():
    pass


ui_commands.add_command(add_ui)
ui_commands.add_command(install_ui)
ui_commands.add_command(uninstall_ui)
ui_commands.add_command(list_uis)



import logging
import sys

from colorama import Fore, Style
from git import Repo

from oarepo_cli.utils import to_python_name
from oarepo_cli.wizard import WizardStep

log = logging.getLogger("step_01_initialize_directory")


class MonorepoDirectoryStep(WizardStep):
    def after_run(self):
        self.data["project_package"] = to_python_name(self.data.project_dir.name)
        p = self.data.project_dir
        if not p.exists():
            print(f"{Fore.BLUE}Creating {Style.RESET_ALL} {p}", file=sys.__stderr__)
            p.mkdir(parents=True)
        if not (p / ".git").exists():
            Repo.init(p)

    def should_run(self):
        return not self.data.project_dir.exists()



from oarepo_cli.model.gen.base import GeneratedFile
from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.package_versions import OAREPO_MODEL_BUILDER_TESTS_VERSION
from oarepo_cli.utils import unique_merger
from oarepo_cli.wizard import RadioStep


class AddTestsWizardStep(ModelWizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            RadioStep(
                "use_tests",
                heading="Do you want to pre-generate some tests?",
                options={"yes": "Yes, I want to use tests", "no": "No"},
            ),
            **kwargs,
        )

    def after_run(self):
        if self.data["use_tests"] != "yes":
            return

        yaml_file: GeneratedFile = self.root.files.get("model.yaml")
        yaml = yaml_file.yaml
        unique_merger.merge(
            yaml,
            {
                "plugins": {
                    "packages": [
                        f"oarepo-model-builder-tests{OAREPO_MODEL_BUILDER_TESTS_VERSION}"
                    ]
                },
            },
        )
        yaml_file.save()

    def should_run(self):
        return True



from typing import List

import click as click

from oarepo_cli.utils import commit_git, with_config

from ...config import MonorepoConfig
from ...site.site_support import SiteSupport
from ...wizard.docker import DockerRunner
from ..model_support import ModelSupport
from .wizard import UnInstallModelWizard


@click.command(
    name="uninstall",
    help="""
Uninstall the model from the current site. Required arguments:
    <name>   ... name of the already existing model""",
)
@click.argument("name", required=True)
@click.argument("site_name", required=False)
@with_config(config_section=lambda name, **kwargs: ["models", name])
def uninstall_model(
    cfg: MonorepoConfig = None,
    no_input=False,
    silent=False,
    step=None,
    steps=False,
    verbose=False,
    site_name=None,
    **kwargs,
):
    commit_git(
        cfg.project_dir,
        f"before-model-uninstall-{cfg.section}",
        f"Committed automatically before model {cfg.section} has been uninstalled",
    )
    site_support = SiteSupport(cfg, site_name)

    model_sites: List[str] = cfg.setdefault("sites", [])

    if site_support.site_name in model_sites:
        model_sites.remove(site_support.site_name)
    cfg.save()

    runner = DockerRunner(cfg, no_input)
    wizard = UnInstallModelWizard(
        runner, model_support=ModelSupport(cfg), site_support=site_support
    )

    if steps:
        wizard.list_steps()
        return

    wizard.run_wizard(
        cfg, no_input=no_input, silent=silent, selected_steps=step, verbose=verbose
    )
    commit_git(
        cfg.project_dir,
        f"after-model-uninstall-{cfg.section}",
        f"Committed automatically after model {cfg.section} has been uninstalled",
    )



import queue
import threading
import traceback
from queue import Queue

from oarepo_cli.develop.config import CONTROL_PIPE
from oarepo_cli.develop.controller import PipeController, TerminalController
from oarepo_cli.develop.runners.docker import DockerDevelopmentRunner
from oarepo_cli.develop.runners.local import LocalDevelopmentRunner
from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.wizard import WizardStep


class DevelopStep(WizardStep):
    def should_run(self):
        return True

    def after_run(self):
        site_support: SiteSupport = self.root.site_support
        if not self.data.running_in_docker and self.data.use_docker:
            runner = DockerDevelopmentRunner(site_support)
        else:
            runner = LocalDevelopmentRunner(site_support)

        if not self.data.running_in_docker:
            controller = TerminalController()
        else:
            controller = PipeController(CONTROL_PIPE)

        control_queue = Queue()
        control_thread = threading.Thread(
            target=lambda: controller.run(control_queue), daemon=True
        )
        control_thread.start()
        self.control_loop(site_support, runner, control_queue)

    def control_loop(self, site_support, runner, control_queue: queue.Queue):
        runner.start()
        try:
            while True:
                try:
                    try:
                        command = control_queue.get(block=True, timeout=10)
                        print(f"Got {command=}")
                    except queue.Empty:
                        continue
                    if not command:
                        continue
                    if command == "stop":
                        runner.stop()
                        break
                    if command == "server":
                        runner.restart_python()
                        continue
                    if command == "ui":
                        runner.restart_ui()
                        continue
                    if command == "build":
                        runner.stop()
                        site_support.rebuild_site(clean=True, build_ui=True)
                        runner.start()
                except InterruptedError:
                    raise
                except Exception:
                    traceback.print_exc()
        except:
            runner.stop()
            return



from oarepo_cli.wizard import Wizard
from oarepo_cli.wizard.docker import DockerRunner

from .steps.uninstall import UnInstallModelStep


class UnInstallModelWizard(Wizard):
    def __init__(self, runner: DockerRunner, *, model_support, site_support):
        self.model_support = model_support
        self.site_support = site_support

        super().__init__(
            *runner.wrap_docker_steps(
                UnInstallModelStep(),
            ),
        )



from oarepo_cli.model.gen.base import GeneratedFile
from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.utils import unique_merger
from oarepo_cli.wizard import InputStep


class AddPIDTypeWizardStep(ModelWizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            InputStep(
                "pid_type",
                prompt="Do you want to specify your own pid type (if empty, it will be autogenerated)?",
                required=False,
            ),
            **kwargs
        )

    def after_run(self):
        if not self.data.get("pid_type"):
            return

        yaml_file: GeneratedFile = self.root.files.get("model.yaml")
        yaml = yaml_file.yaml
        unique_merger.merge(
            yaml,
            {
                "record": {"pid": {"type": self.data["pid_type"]}},
            },
        )
        yaml_file.save()

    def should_run(self):
        return True



import json
import tempfile

from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import WizardStep


class CheckSearchStep(SiteWizardStepMixin, WizardStep):
    def after_run(self):
        self.site_support.call_invenio("oarepo", "index", "init")
        # TODO: add option to reindex data to tooling

    def should_run(self):
        with tempfile.NamedTemporaryFile(suffix=".json") as f:
            self.site_support.call_invenio(
                "oarepo", "check", f.name, raise_exception=True, grab_stdout=True
            )
            f.seek(0)
            data = json.load(f)
            return data["opensearch"] != "ok"



"""Additional views."""

from flask import Blueprint


#
# Registration
#
def create_blueprint(app):
    """Register blueprint routes on app."""
    blueprint = Blueprint(
        "{{cookiecutter.package_name}}",
        __name__,
        template_folder="./templates",
    )

    # Add URL rules
    return blueprint



import click as click

from .add.cli import add_model
from .compile.cli import compile_model
from .install.cli import install_model
from .list import list_models
from .uninstall.cli import uninstall_model


@click.group(
    help="Model-related tools (add model, compile, install, load and dump data)",
    name="model",
)
def model_commands():
    pass


model_commands.add_command(add_model)
model_commands.add_command(compile_model)
model_commands.add_command(install_model)
model_commands.add_command(uninstall_model)
model_commands.add_command(list_models)



from functools import cached_property


class SiteWizardStepMixin:
    @cached_property
    def site_support(self):
        return self.root.site_support

    @property
    def site_dir(self):
        return self.site_support.site_dir

    def get_invenio_configuration(self, *keys):
        return self.site_support.get_invenio_configuration(*keys)



# Copyright (c) 2022 Miroslav Bauer
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import re

from invenio_base.utils import obj_or_import_string
from {{cookiecutter.app_package}} import config as config


class {{cookiecutter.ext_name}}:
    def __init__(self, app=None):
        if app:
            self.init_config(app)
            self.init_app(app)
            self.init_resource(app)

    def init_app(self, app):
        app.extensions["{{cookiecutter.app_package}}"] = self

    def init_resource(self, app):
        """Initialize vocabulary resources."""
        self.resource = obj_or_import_string(app.config["{{cookiecutter.app_package | upper}}_RESOURCE"])(
            config=obj_or_import_string(app.config["{{cookiecutter.app_package | upper}}_RESOURCE_CONFIG"])(),
        )

    def init_config(self, app):
        """Initialize configuration."""
        for identifier in dir(config):
            if re.match("^[A-Z_]*$", identifier) and not identifier.startswith("_"):
                app.config.setdefault(identifier, getattr(config, identifier))




import click as click

from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import commit_git, with_config
from oarepo_cli.wizard.docker import DockerRunner

from .wizard import AddUIWizard


@click.command(
    name="add",
    help="""Generate a new UI. Required arguments:
    <name>   ... name of the ui. The recommended pattern for it is <modelname>-ui
    """,
)
@click.argument("name")
@with_config(config_section=lambda name, **kwargs: ["ui", name])
def add_ui(
    cfg=None,
    step=None,
    no_input=False,
    silent=False,
    verbose=False,
    steps=False,
    **kwargs,
):
    commit_git(
        cfg.project_dir,
        f"before-ui-add-{cfg.section}",
        f"Committed automatically before ui {cfg.section} has been added",
    )

    site_support = SiteSupport(cfg)

    runner = DockerRunner(cfg, no_input)

    wizard = AddUIWizard(runner, site_support=site_support)

    if steps:
        wizard.list_steps()
        return

    wizard.run_wizard(
        cfg, selected_steps=step, no_input=no_input, silent=silent, verbose=verbose
    )
    commit_git(
        cfg.project_dir,
        f"after-ui-add-{cfg.section}",
        f"Committed automatically after ui {cfg.section} has been added",
    )



from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import WizardStep


class CompileGUIStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
Now I'll compile Invenio GUI.

Note that this can take a lot of time as UI dependencies
will be downloaded and installed and UI will be compiled.
            """,
            **kwargs,
        )

    def after_run(self):
        self.site_support.build_ui()

    def should_run(self):
        manifest_file = self._manifest_file
        return not manifest_file.exists()

    @property
    def _manifest_file(self):
        manifest_file = (
            self.site_support.invenio_instance_path
            / "static"
            / "dist"
            / "manifest.json"
        )

        return manifest_file



import os
import subprocess
import sys
import time
from pathlib import Path

from oarepo_cli.kill import kill
from oarepo_cli.site.site_support import SiteSupport


class LocalDevelopmentRunner:
    def __init__(self, site_support: SiteSupport):
        self.site_support: SiteSupport = site_support
        self.server_handle = None
        self.ui_handle = None

    def start(self):
        self.start_server()
        self.start_ui()

    def stop(self):
        self.stop_server()
        self.stop_ui()

    def restart_python(self):
        self.stop_server()
        self.start_server()

    def restart_ui(self):
        self.stop_ui()
        self.start_ui()

    @property
    def nrp_cli(self):
        return Path(sys.argv[0]).resolve()

    def start_server(self):
        print("Starting server")
        self.server_handle = subprocess.Popen(
            [
                self.nrp_cli,
                "run",
                "--site",
                self.site_support.site_name,
                "--use-docker"
                if self.site_support.config.use_docker
                else "--outside-docker",
            ],
            env={
                "INVENIO_TEMPLATES_AUTO_RELOAD": "1",
                "FLASK_DEBUG": "1",
                **os.environ,
            },
            stdin=subprocess.DEVNULL,
            cwd=self.site_support.config.project_dir,
        )

    def stop_server(self):
        print("Stopping server")
        self.stop_handle(self.server_handle)
        self.server_handle = None

    def start_ui(self):
        print("Starting file watcher")
        self.ui_handle = subprocess.Popen(
            [
                self.nrp_cli,
                "ui-watch",
                "--site",
                self.site_support.site_name,
                "--use-docker"
                if self.site_support.config.use_docker
                else "--outside-docker",
                "--run-ui",
            ],
            stdin=subprocess.DEVNULL,
            cwd=self.site_support.config.project_dir,
        )
        time.sleep(5)

    def stop_ui(self):
        print("Stopping file watcher")
        self.stop_handle(self.ui_handle)
        self.ui_handle = None

    @staticmethod
    def stop_handle(handle):
        if handle and handle.returncode is None:
            kill(handle.pid)



import click as click

from .add.cli import add_local
from .remove.cli import remove_local


@click.group(help="Support for local packages", name="local")
def local_commands():
    pass


local_commands.add_command(add_local)
local_commands.add_command(remove_local)



from oarepo_cli.format.steps.base import BaseFormatStep


class FormatJinjaStep(BaseFormatStep):
    def format_paths(self, model_paths, ui_paths, local_paths, site_paths):
        self.format_jinja(ui_paths + local_paths, exclude=[])

    def format_jinja(self, dirs, exclude):
        # look for 'templates' folder inside the paths
        from djlint import Config, process

        for f in self.find_files_with_extensions(dirs, "html"):
            if not "templates" in f.parts:
                continue
            config = Config(src=f, profile="jinja", reformat=True)
            process(config, f)



import shutil

from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.wizard import WizardStep


class CheckDependenciesStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, clean=False):
        super().__init__()
        self.clean = clean

    def after_run(self):
        self.site_support.build_dependencies()

    def should_run(self):
        return self.clean or not (self.site_support.site_dir / 'requirements.txt').exists()



import errno
import os
import queue
import select
import sys

# Controller is a class that gets commands from user - it may be a terminal controller, where the input
# is taken from stdin, or pipe controller, where the input is taken from a unix pipe
# The input is then put to a blocking queue
# The controller is run as a daemon thread from the main loop.


class TerminalController:
    def run(self, queue: queue.Queue):
        running = True
        try:
            while running:
                cmd = self.input_with_timeout(60)
                if not cmd:
                    continue
                if cmd == "stop":
                    running = False
                queue.put(cmd)
        except InterruptedError:
            queue.put("stop")

    def input_with_timeout(self, timeout):
        print("=======================================================================")
        print()
        print("Type: ")
        print()
        print("    server <enter>    --- restart server")
        print("    ui <enter>        --- restart ui watcher")
        print("    build <enter>     --- stop server and watcher, ")
        print("                          call ui build, then start again")
        print("    stop <enter>      --- stop the server and ui and exit")
        print()
        i, o, e = select.select([sys.stdin], [], [], timeout)

        if i:
            return sys.stdin.readline().strip()


class PipeController:
    def __init__(self, pipe_path):
        self.pipe_path = pipe_path
        try:
            os.mkfifo(pipe_path)
        except OSError as oe:
            if oe.errno != errno.EEXIST:
                raise

    def run(self, queue: queue.Queue):
        running = True
        while running:
            with open(self.pipe_path) as fd:
                cmd = fd.read(1024)
                if not cmd:
                    continue
                cmd = cmd.strip()
                if cmd == "stop":
                    running = False
                queue.put(cmd)
                if not running:
                    break



from oarepo_cli.model.gen.base import GeneratedFile
from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.package_versions import OAREPO_MODEL_BUILDER_CF_VERSION
from oarepo_cli.utils import unique_merger
from oarepo_cli.wizard import RadioStep


class AddCustomFieldsWizardStep(ModelWizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            RadioStep(
                "use_custom_fields",
                heading="Do you want your model to have configurable deployment custom fields?",
                options={
                    "yes": "Yes, I want that extensibility",
                    "no": "No, the metadata schema should be fixed",
                },
            ),
            **kwargs,
        )

    def after_run(self):
        if self.data["use_custom_fields"] != "yes":
            return

        yaml_file: GeneratedFile = self.root.files.get("model.yaml")
        yaml = yaml_file.yaml
        unique_merger.merge(
            yaml,
            {
                "record": {"use": ["./custom_fields.yaml"]},
                "plugins": {
                    "packages": [
                        f"oarepo-model-builder-cf{OAREPO_MODEL_BUILDER_CF_VERSION}"
                    ]
                },
            },
        )
        yaml_file.save()

        yaml_file: GeneratedFile = self.root.files.get("custom_fields.yaml")
        yaml = yaml_file.yaml

        cf = unique_merger.merge(yaml, {"custom-fields": []})
        if not yaml.ca.comment:
            yaml.yaml_set_comment_before_after_key(
                key="custom-fields",
                before="""
    TODO: add your expandable fields definitions for the base record here
    example:
    custom-fields:
    - element: custom_fields
      config: TEST_CUSTOM_FIELDS
    - config: INLINE_CUSTOM_FIELDS        
            """.strip(),
            )
        yaml_file.save()

    def should_run(self):
        return True



import time

import click
import pika
import psycopg
import redis
from minio import Minio
from opensearchpy import OpenSearch

from oarepo_cli.site.mixins import SiteWizardStepMixin
from oarepo_cli.utils import run_cmdline
from oarepo_cli.wizard import WizardStep


class StartContainersStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
I'm going to start docker containers (database, opensearch, message queue, cache etc.).
If this step fails, please fix the problem and run the wizard again.            
            """,
            **kwargs,
        )

    def after_run(self):
        run_cmdline(
            "docker",
            "compose",
            "up",
            "-d",
            "cache",
            "db",
            "mq",
            "search",
            "opensearch-dashboards",
            "pgadmin",
            "s3",
            cwd=self.site_dir,
            with_tty=False,
        )
        self._check_containers_running(False)

    def _check_containers_running(self, check_only):
        def retry(fn, tries=10, timeout=1):
            click.secho(f"Calling {fn.__name__}", fg="yellow")
            for i in range(tries):
                try:
                    fn()
                    click.secho(f"  ... alive", fg="green")
                    return
                except InterruptedError:
                    raise
                except Exception as e:
                    if check_only:
                        raise
                    self.vprint(e)
                    if i == tries - 1:
                        click.secho(f" ... failed", fg="red")
                        raise
                    click.secho(
                        f" ... not yet ready, sleeping for {int(timeout)} seconds",
                        fg="yellow",
                    )
                    time.sleep(int(timeout))
                    nt = timeout * 1.5
                    if int(nt) == int(timeout):
                        timeout = timeout + 1
                    else:
                        timeout = nt

        try:
            retry(self.check_redis)
            retry(self.check_db)
            retry(self.check_mq)
            retry(self.check_s3)
            retry(self.check_search)
            return True
        except InterruptedError:
            raise
        except:
            if check_only:
                return False
            raise

    def check_redis(self):
        host, port = self.get_invenio_configuration(
            "INVENIO_REDIS_HOST", "INVENIO_REDIS_PORT"
        )
        pool = redis.ConnectionPool(host=host, port=port, db=0)
        r = redis.Redis(connection_pool=pool)
        r.keys("blahblahblah")  # fails if there is no connection
        pool.disconnect()

    def check_db(self):
        host, port, user, password, dbname = self.get_invenio_configuration(
            "INVENIO_DATABASE_HOST",
            "INVENIO_DATABASE_PORT",
            "INVENIO_DATABASE_USER",
            "INVENIO_DATABASE_PASSWORD",
            "INVENIO_DATABASE_DBNAME",
        )

        with psycopg.connect(
            dbname=dbname, host=host, port=port, user=user, password=password
        ) as conn:
            assert conn.execute("SELECT 1").fetchone()[0] == 1

    def check_mq(self):
        host, port = self.get_invenio_configuration(
            "INVENIO_RABBIT_HOST", "INVENIO_RABBIT_PORT"
        )
        connection = pika.BlockingConnection(
            pika.ConnectionParameters(host=host, port=port)
        )
        channel = connection.channel()
        connection.process_data_events(2)
        assert connection.is_open
        connection.close()

    def check_s3(self):
        host, port, access_key, secret_key = self.get_invenio_configuration(
            "INVENIO_S3_HOST",
            "INVENIO_S3_PORT",
            "INVENIO_S3_ACCESS_KEY",
            "INVENIO_S3_SECRET_KEY",
        )

        client = Minio(
            f"{host}:{port}",
            access_key=access_key,
            secret_key=secret_key,
            secure=False,
        )
        client.list_buckets()

    def check_search(self):
        (
            host,
            port,
            prefix,
            use_ssl,
            verify_certs,
            ssl_assert_hostname,
            ssl_show_warn,
        ) = self.get_invenio_configuration(
            "INVENIO_OPENSEARCH_HOST",
            "INVENIO_OPENSEARCH_PORT",
            "INVENIO_SEARCH_INDEX_PREFIX",
            "INVENIO_OPENSEARCH_USE_SSL",
            "INVENIO_OPENSEARCH_VERIFY_CERTS",
            "INVENIO_OPENSEARCH_ASSERT_HOSTNAME",
            "INVENIO_OPENSEARCH_SHOW_WARN",
        )
        try:
            ca_certs = self.get_invenio_configuration(
                "INVENIO_OPENSEARCH_CA_CERTS_PATH"
            )[0]
        except:
            ca_certs = None
        client = OpenSearch(
            hosts=[{"host": host, "port": port}],
            use_ssl=use_ssl,
            verify_certs=verify_certs,
            ssl_assert_hostname=ssl_assert_hostname,
            ssl_show_warn=ssl_show_warn,
            ca_certs=ca_certs,
        )
        info = client.info(pretty=True)
        self.vprint(info)

    def should_run(self):
        return not self._check_containers_running(True)



import json
import re
from pathlib import Path
from typing import Any, List

from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import ProjectWizardMixin, SiteMixin
from oarepo_cli.wizard import WizardStep

from ..mixins import AssociatedModelMixin


def replace_non_variable_signs(x):
    return f"__{ord(x.group())}__"


class CreateJinjaStep(SiteMixin, AssociatedModelMixin, ProjectWizardMixin, WizardStep):
    def should_run(self):
        return True

    def after_run(self):
        (
            model_description,
            model_path,
            model_package,
            model_config,
        ) = self.get_model_definition()
        # get the UI definition
        ui_definition_path = model_path / model_package / "models" / "ui.json"
        ui_definition = json.loads(ui_definition_path.read_text())

        site = SiteSupport(self.data)
        renderers_json = site.call_invenio(
            "oarepo",
            "ui",
            "renderers",
            "--json",
            grab_stdout=True,
        )
        renderers = [x["renderer"] for x in json.loads(renderers_json)]

        template, macro_definitions = self.generate_main(ui_definition)
        if macro_definitions:
            macros = "\n".join(
                self.generate_macro_definitions(macro_definitions, set(renderers))
            )
        else:
            macros = None

        # save template and macros
        ui_dir = self.data.project_dir / self.data.config["ui_dir"]
        main_jinja_path = (
            ui_dir
            / self.data.config["cookiecutter_app_package"]
            / "templates"
            / "semantic-ui"
            / self.data.config["cookiecutter_app_package"]
            / "main.html"
        )
        template = main_jinja_path.read_text() + "\n\n" + template
        main_jinja_path.write_text(template)

        macros_jinja_path: Path = (
            ui_dir
            / self.data.config["cookiecutter_app_package"]
            / "templates"
            / "semantic-ui"
            / "oarepo_ui"
            / "components"
            / "100-macros.html"
        )
        macros_jinja_path.parent.mkdir(exist_ok=True, parents=True)
        macros_jinja_path.write_text(macros or "")

    def _select(self, fields, *keys):
        for k in keys:
            if k in fields:
                return k, fields.pop(k)
        return None, None

    def generate_main(self, ui):
        macro_definitions = []
        template = []
        fields = ui["children"]
        if "metadata" in fields:
            md = fields.pop("metadata")
            fields.update(
                {f"metadata.{k}": v for k, v in md.get("children", {}).items()}
            )
        title_key, title = self._select(fields, "title", "metadata.title")
        divider = False
        if title_key:
            template.append(f'<h1>{{%- value "{title_key}" -%}}</h1>')
            macro_definitions.append(title)
            divider = True
        creator_key, creator = self._select(fields, "creator", "metadata.creator")
        if creator_key:
            template.append(
                f'<div class="creator">{{%- value "{creator_key}" -%}}</div>'
            )
            macro_definitions.append(creator)
            divider = True
        if divider:
            template.append('<hr class="divider"/>')
        template.append('<dl class="detail-fields">')
        for fld_key, fld in sorted(fields.items()):
            template.append(f'{{%- field "{fld_key}" -%}}')
            macro_definitions.append(fld)
        template.append("</dl>")

        return "\n".join(template), macro_definitions

    def generate_macro_definitions(
        self, macro_definitions: List[Any], processed_components
    ):
        for definition in macro_definitions:
            if not definition.get("detail"):
                continue
            component = re.sub(r"\W", replace_non_variable_signs, definition["detail"])

            if component in processed_components:
                _, children = self._generate_macro_children(definition)
            else:
                processed_components.add(component)

                children_def, children = self._generate_macro_children(definition)
                if children_def:
                    yield f"\n\n{{%- macro render_{component}(arg) -%}}\n<dl class='detail-subfields'>\n{children_def}\n</dl>\n{{%- endmacro -%}}"
                else:
                    yield f"\n\n{{%- macro render_{component}(arg) -%}}{'{{'}arg{'}}'}{{%- endmacro -%}}"

            yield from self.generate_macro_definitions(children, processed_components)

    def _generate_macro_children(self, definition):
        # for array members, do not return fields as array macro is built-in
        if "child" in definition:
            return "", [definition["child"]]
        if "children" not in definition:
            return "", []
        fields = []
        children = []
        for c_key, cdef in definition["children"].items():
            fields.append(f'{{%- field "{c_key}" -%}}')
            children.append(cdef)
        return "\n".join(fields), children



import click

from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import with_config
from oarepo_cli.watch.wizard import WatcherWizard
from oarepo_cli.wizard.docker import DockerRunner


@click.command(
    name="ui-watch",
    hidden=True,
    help="Internal action called inside the development docker. " "It watches the ",
)
@with_config()
@click.option("--run-ui", is_flag=True)
@click.option("--site")
def watch_command(
    run_ui,
    site,
    cfg=None,
    step=None,
    no_input=False,
    silent=False,
    verbose=False,
    steps=False,
    **kwargs,
):
    site_support = SiteSupport(cfg, site)

    runner = DockerRunner(cfg, no_input)
    wizard = WatcherWizard(runner, site_support, run_ui)
    if steps:
        wizard.list_steps()
        return

    wizard.run_wizard(
        cfg, selected_steps=step, no_input=no_input, silent=silent, verbose=verbose
    )



from pathlib import Path

from oarepo_cli.config import MonorepoConfig


class ModelSupport:
    def __init__(self, config: MonorepoConfig, model_section=None):
        self.config = config
        models = config.whole_data.get("models", {})

        if config.section_path[0] == "models":
            self.model = config
            self.model_name = config.section_path[-1]
            return
        elif not model_section:
            if len(models) == 1:
                model_section = next(iter(models.keys()))
            else:
                raise RuntimeError(
                    "none or more models, please specify the model on commandline"
                )
        self.model = models[model_section]
        self.model_name = model_section

    @property
    def model_dir(self):
        return Path(self.config.project_dir) / self.config["model_dir"]



import click

from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.utils import commit_git, with_config
from oarepo_cli.wizard.docker import DockerRunner

from .wizard import FormatWizard


@click.command(
    name="format",
    help="""Format all source files inside the project""",
)
@with_config()
def format_sources_command(
    cfg=None,
    no_input=False,
    silent=False,
    step=None,
    verbose=False,
    steps=False,
    **kwargs,
):
    commit_git(
        cfg.project_dir,
        f"before-file-format-{cfg.section}",
        f"Committed automatically before file formatting",
    )
    site_support = SiteSupport(cfg)

    runner = DockerRunner(cfg, no_input)
    format_wizard = FormatWizard(runner, site_support)
    if steps:
        format_wizard.list_steps()
        return

    format_wizard.run_wizard(
        cfg, no_input=no_input, silent=silent, selected_steps=step, verbose=verbose
    )
    commit_git(
        cfg.project_dir,
        f"after-file-format-{cfg.section}",
        f"Committed automatically after files have been formatted",
    )



from oarepo_cli.wizard import Wizard

from ...wizard.docker import DockerRunner
from ..site_support import SiteSupport
from .steps.check_requirements import CheckRequirementsStep
from .steps.compile_gui import CompileGUIStep
from .steps.init_database import InitDatabaseStep
from .steps.init_files import InitFilesStep
from .steps.install_invenio import InstallInvenioStep
from .steps.install_site import InstallSiteStep
from .steps.link_env import LinkEnvStep
from .steps.next_steps import NextStepsStep
from .steps.resolve_dependencies import ResolveDependenciesStep
from .steps.start_containers import StartContainersStep


class AddSiteWizard(Wizard):
    def __init__(self, runner: DockerRunner, site_support: SiteSupport):
        self.site_support = site_support
        steps = []
        steps.extend(
            runner.wrap_docker_steps(
                InstallSiteStep(), in_compose=False, interactive=True
            )  # can run in plain docker
        )
        steps.append(LinkEnvStep())  # runs in userspace
        steps.extend(
            runner.wrap_docker_steps(
                CheckRequirementsStep(),  # can run in docker compose
            )
        )
        steps.append(StartContainersStep())  # runs in userspace
        steps.extend(
            runner.wrap_docker_steps(
                ResolveDependenciesStep(),  # can run in docker compose
                InstallInvenioStep(),  # can run in docker compose
                CompileGUIStep(),  # can run in docker compose
                InitDatabaseStep(),  # can run in docker compose
                InitFilesStep(),  # can run in docker compose
            )
        )
        steps.extend([NextStepsStep()])  # runs in userspace
        super().__init__(*steps)



from oarepo_cli.model.gen.base import GeneratedFile
from oarepo_cli.model.utils import ModelWizardStep
from oarepo_cli.package_versions import (
    NR_METADATA_VERSION,
    NR_OAIPMH_HARVESTERS_VERSION,
    NR_VOCABULARIES_VERSION,
    OAREPO_MODEL_BUILDER_NR_VERSION,
)
from oarepo_cli.utils import unique_merger
from oarepo_cli.wizard import RadioStep


class AddDocsBaseWizardStep(ModelWizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            RadioStep(
                "use_docs",
                heading="Should I base the model on nr-docs?",
                options={"yes": "Yes", "no": "No"},
            ),
            **kwargs,
        )

    def after_run(self):
        if self.data["use_docs"] != "yes":
            return
        self.data["use_vocabularies"] = "yes"
        self.data["use_nr_vocabularies"] = "yes"
        self.data["use_metadata"] = "yes"
        yaml_file: GeneratedFile = self.root.files.get("model.yaml")
        yaml = yaml_file.yaml
        unique_merger.merge(
            yaml,
            {
                "record": {
                    "extend": "nr-documents#DocumentModel",
                },
                "plugins": {
                    "builder": {"disable": ["script_sample_data"]},
                    "packages": [
                        f"oarepo-model-builder-nr{OAREPO_MODEL_BUILDER_NR_VERSION}"
                    ],
                },
                "runtime-dependencies": {
                    "nr-metadata": NR_METADATA_VERSION,
                    "nr-oaipmh-harvesters": NR_OAIPMH_HARVESTERS_VERSION,
                    "nr-vocabularies": NR_VOCABULARIES_VERSION,
                },
                "settings": {
                    "i18n-languages": ["cs", "en"],
                    "supported-langs": {
                        "cs": {
                            "keyword": {"type": "keyword"},
                            "text": {"analyzer": "czech"},
                        },
                        "en": {"text": {}},
                    },
                },
            },
        )
        yaml_file.save()

    def should_run(self):
        return True



import os

from oarepo_cli.site.site_support import SiteSupport
from oarepo_cli.wizard import WizardStep


class RunServerStep(WizardStep):
    def should_run(self):
        return True

    def after_run(self):
        site_support: SiteSupport = self.root.site_support
        site_support.call_invenio(
            "run",
            "--cert",
            "docker/nginx/test.crt",
            "--key",
            "docker/nginx/test.key",
            "-h",
            os.environ.get("INVENIO_UI_HOST", "127.0.0.1"),
        )



from oarepo_cli.ui.install.steps.build_assets_step import BuildAssetsUIStep
from oarepo_cli.ui.install.steps.install_step import InstallUIStep
from oarepo_cli.utils import ProjectWizardMixin
from oarepo_cli.wizard import Wizard
from oarepo_cli.wizard.docker import DockerRunner


class InstallWizard(ProjectWizardMixin, Wizard):
    def __init__(self, runner: DockerRunner, *, site_support):
        self.site_support = site_support
        super().__init__(
            *runner.wrap_docker_steps(InstallUIStep(), BuildAssetsUIStep())
        )



import click

from oarepo_cli.config import MonorepoConfig
from oarepo_cli.site.site_support import SiteSupport
from .wizard import BuildWizard
from oarepo_cli.utils import with_config
from oarepo_cli.wizard.docker import DockerRunner


@click.command(name="build")
@with_config()
@click.option("--site")
@click.option("--production/--development")
@click.pass_context
def build_command(
    ctx,
    cfg: MonorepoConfig = None,
    step=None,
    no_input=False,
    silent=False,
    verbose=False,
    steps=False,
    site=None,
        production=None,
    **kwargs,
):
    site_support = SiteSupport(cfg, site)

    runner = DockerRunner(cfg, no_input=True)
    wizard = BuildWizard(runner, site_support, production)

    if site:
        cfg = cfg.clone(["sites", site])
        wizard.site_support = SiteSupport(cfg, site)

    if steps:
        wizard.list_steps()
        return

    wizard.run_wizard(
        cfg, selected_steps=step, no_input=no_input, silent=silent, verbose=verbose
    )



from pathlib import Path

from oarepo_cli.wizard import WizardStep


class BaseFormatStep(WizardStep):
    def should_run(self):
        return True

    def after_run(self):
        pd = self.data.project_dir
        model_paths = [
            pd / "models" / model_name / model["model_package"]
            for model_name, model in self.data.whole_data.get("models", {}).items()
            if (pd / "models" / model_name / model["model_package"]).exists()
        ]
        ui_paths = [
            pd
            / "ui"
            / ui_name
            / ui["cookiecutter_app_package"]  # TODO: better name here?
            for ui_name, ui in self.data.whole_data.get("ui", {}).items()
            if (pd / "ui" / ui_name / ui["cookiecutter_app_package"]).exists()
        ]
        local_paths = [
            pd / "local" / local_name
            for local_name in self.data.whole_data.get("local", {})
            if (pd / "local" / local_name).exists()
        ]
        site_paths = [
            pd / "sites" / site_name
            for site_name in self.data.whole_data.get("sites", {})
            if (pd / "sites" / site_name).exists()
        ]
        self.format_paths(model_paths, ui_paths, local_paths, site_paths)

    def format_paths(self, model_paths, ui_paths, local_paths, site_paths):
        pass

    def find_files_with_extensions(self, dirs, *extensions):
        for d in dirs:
            for f in Path(d).glob("**/*"):
                if f.suffix[1:] in extensions:
                    yield f
